import pytest

import asdf
import astropy
from astropy import units as u
from astropy.units import equivalencies as eq
from astropy.cosmology import Planck15
from packaging.version import Version


def create_equivalencies():
    result = [
        eq.plate_scale(.3 * u.deg/u.mm), eq.pixel_scale(.5 * u.deg/u.pix),
        eq.spectral_density(350 * u.nm, factor=2),
        eq.spectral_density(350 * u.nm), eq.spectral(),
        eq.brightness_temperature(500 * u.GHz),
        eq.brightness_temperature(500 * u.GHz, beam_area=23 * u.sr),
        eq.with_H0(), eq.temperature_energy(), eq.temperature(),
        eq.thermodynamic_temperature(300 * u.Hz),
        eq.thermodynamic_temperature(140 * u.GHz, Planck15.Tcmb0),
        eq.beam_angular_area(3 * u.sr), eq.mass_energy(),
        eq.molar_mass_amu(), eq.doppler_relativistic(2 * u.m),
        eq.doppler_optical(2 * u.nm), eq.doppler_radio(2 * u.Hz),
        eq.parallax(), eq.logarithmic(), eq.dimensionless_angles(),
        eq.spectral() + eq.temperature(),
        (eq.spectral_density(35 * u.nm) +
            eq.brightness_temperature(5 * u.Hz, beam_area=2 * u.sr)),
        (eq.spectral() + eq.spectral_density(35 * u.nm) +
            eq.brightness_temperature(5 * u.Hz, beam_area=2 * u.sr)),
    ]

    if Version(astropy.__version__) >= Version("4.1"):
        result.append(eq.pixel_scale(100. * u.pix/u.cm))

    return result


@pytest.mark.parametrize("equivalency", create_equivalencies())
def test_serialization(equivalency, tmp_path):
    file_path = tmp_path / "test.asdf"
    with asdf.AsdfFile() as af:
        af["equivalency"] = equivalency
        af.write_to(file_path)

    with asdf.open(file_path) as af:
        assert af["equivalency"] == equivalency



from packaging.version import parse as parse_version

from .core import TransformConverterBase, parameter_to_value

from ..helpers import parse_tag_version


class ConstantConverter(TransformConverterBase):
    """
    ASDF support for serializing the Const1D and Const2D models.
    """
    # The 'dimensions' property was added in 1.4.0,
    # previously all values were 1D.
    _2D_MIN_VERSION = parse_version("1.4.0")

    tags = ["tag:stsci.edu:asdf/transform/constant-*"]

    types = [
        "astropy.modeling.functional_models.Const1D",
        "astropy.modeling.functional_models.Const2D",
    ]

    def to_yaml_tree_transform(self, model, tag, ctx):
        from astropy.modeling.functional_models import Const1D, Const2D

        if parse_tag_version(tag) < self._2D_MIN_VERSION:
            if not isinstance(model, Const1D):
                raise TypeError(
                    f"{tag} does not support models with > 1 dimension")
            return {
                "value": parameter_to_value(model.amplitude)
            }
        else:
            if isinstance(model, Const1D):
                dimension = 1
            elif isinstance(model, Const2D):
                dimension = 2
            return {
                "value": parameter_to_value(model.amplitude),
                "dimensions": dimension
            }

    def from_yaml_tree_transform(self, node, tag, ctx):
        from astropy.modeling.functional_models import Const1D, Const2D

        if parse_tag_version(tag) < self._2D_MIN_VERSION:
            return Const1D(node["value"])
        elif node["dimensions"] == 1:
            return Const1D(node["value"])
        elif node["dimensions"] == 2:
            return Const2D(node["value"])



from .core import TransformConverterBase


__all__ = ["TabularConverter"]


class TabularConverter(TransformConverterBase):
    """
    ASDF support for serializing tabular models.
    """
    tags = ["tag:stsci.edu:asdf/transform/tabular-*"]

    types = [
        "astropy.modeling.tabular.Tabular1D",
        "astropy.modeling.tabular.Tabular2D",
    ]

    def to_yaml_tree_transform(self, model, tag, ctx):
        node = {}
        if model.fill_value is not None:
            node["fill_value"] = model.fill_value
        node["lookup_table"] = model.lookup_table
        node["points"] = [p for p in model.points]
        node["method"] = str(model.method)
        node["bounds_error"] = model.bounds_error

        return node

    def from_yaml_tree_transform(self, node, tag, ctx):
        from astropy.modeling import tabular

        lookup_table = node.pop("lookup_table")
        dim = lookup_table.ndim
        fill_value = node.pop("fill_value", None)
        if dim == 1:
            points = (node["points"][0],)
            model = tabular.Tabular1D(points=points, lookup_table=lookup_table,
                                      method=node["method"], bounds_error=node["bounds_error"],
                                      fill_value=fill_value)
        elif dim == 2:
            points = tuple(node["points"])
            model = tabular.Tabular2D(points=points, lookup_table=lookup_table,
                                      method=node["method"], bounds_error=node["bounds_error"],
                                      fill_value=fill_value)
        else:
            raise NotImplementedError("tabular models with ndim > 2 are not supported ")

        return model



from ..unit.quantity import QuantityConverter


class AngleConverter(QuantityConverter):
    tags = ["tag:astropy.org:astropy/coordinates/angle-*"]

    types = ["astropy.coordinates.angles.Angle"]

    def from_yaml_tree(self, node, tag, ctx):
        from astropy.coordinates.angles import Angle
        return Angle(super().from_yaml_tree(node, tag, ctx))


class LatitudeConverter(QuantityConverter):
    tags = ["tag:astropy.org:astropy/coordinates/latitude-*"]

    types = ["astropy.coordinates.angles.Latitude"]

    def from_yaml_tree(self, node, tag, ctx):
        from astropy.coordinates.angles import Latitude
        return Latitude(super().from_yaml_tree(node, tag, ctx))


class LongitudeConverter(QuantityConverter):
    tags = ["tag:astropy.org:astropy/coordinates/longitude-*"]

    types = ["astropy.coordinates.angles.Longitude"]

    def to_yaml_tree(self, obj, tag, ctx):
        tree = super().to_yaml_tree(obj, tag, ctx)
        tree["wrap_angle"] = obj.wrap_angle
        return tree

    def from_yaml_tree(self, node, tag, ctx):
        from astropy.coordinates.angles import Longitude
        return Longitude(super().from_yaml_tree(node, tag, ctx), wrap_angle=node["wrap_angle"])



import pytest

import asdf
from astropy.coordinates import EarthLocation
from astropy.time import Time
from astropy import units as u
import numpy as np

from asdf_astropy.testing.helpers import assert_time_equal


def create_times():
    return [
        Time(np.arange(100), format="unix"),
        Time(
            [1, 2],
            location=EarthLocation(x=[1, 2] * u.m, y=[3, 4] * u.m, z=[5, 6] * u.m),
            format="cxcsec",
        ),
        Time(
            "J2000.000",
            location=EarthLocation(x=6378100 * u.m, y=0 * u.m, z=0 * u.m),
            format="jyear_str",
        ),
        Time("2000-01-01T00:00:00.000"),
        Time(["2001-01-02T12:34:56", "2001-02-03T00:01:02"]),
    ]


@pytest.mark.parametrize("time", create_times())
@pytest.mark.parametrize("version", asdf.versioning.supported_versions)
def test_serialization(time, version, tmp_path):
    file_path = tmp_path / "test.asdf"
    with asdf.AsdfFile(version=version) as af:
        af["time"] = time
        af.write_to(file_path)

    with asdf.open(file_path) as af:
        assert_time_equal(af["time"], time)



import asdf
import pytest
import numpy as np
import itertools
from packaging.version import Version

import astropy
from astropy.modeling import models as astropy_models
import astropy.modeling
from astropy import units as u
from numpy.testing import assert_array_equal
from asdf.tests.helpers import yaml_to_asdf

from asdf_astropy import integration

try:
    from astropy.modeling.models import UnitsMapping
    HAS_NO_UNITS_MAPPING = False
except ImportError:
    HAS_NO_UNITS_MAPPING = True

if Version(astropy.__version__) < Version("4.1"):
    ASTROPY_LT_41 = True
else:
    ASTROPY_LT_41 = False


def assert_model_roundtrip(model, tmpdir, version=None):
    """
    Assert that a model can be written to an ASDF file and read back
    in without losing any of its essential properties.
    """
    path = str(tmpdir / "test.asdf")

    with asdf.AsdfFile({"model": model}, version=version) as af:
        af.write_to(path)

    with asdf.open(path) as af:
        assert_models_equal(model, af["model"])
        return af["model"]


def assert_models_equal(a, b):
    """
    Assert that two model instances are equivalent.
    """
    if a is None and b is None:
        return

    assert a.__class__ == b.__class__

    assert a.name == b.name
    assert a.inputs == b.inputs
    assert a.input_units == b.input_units
    assert a.outputs == b.outputs
    assert a.input_units_allow_dimensionless == b.input_units_allow_dimensionless

    for i in a.inputs:
        if a.input_units_equivalencies is None:
            a_equiv = None
        else:
            a_equiv = a.input_units_equivalencies.get(i)

        if b.input_units_equivalencies is None:
            b_equiv = None
        else:
            b_equiv = b.input_units_equivalencies.get(i, None)

        assert a_equiv == b_equiv

    assert_array_equal(a.parameters, b.parameters)

    try:
        a_bounding_box = a.bounding_box
    except NotImplementedError:
        a_bounding_box = None

    try:
        b_bounding_box = b.bounding_box
    except NotImplementedError:
        b_bounding_box = None

    assert a_bounding_box == b_bounding_box

    assert a.fixed == b.fixed
    assert a.bounds == b.bounds

    assert_models_equal(a._user_inverse, b._user_inverse)


def create_single_models():
    model_with_bounding_box = astropy_models.Shift(10)
    model_with_bounding_box.bounding_box = ((1, 7),)

    model_with_user_inverse = astropy_models.Shift(10)
    model_with_user_inverse.inverse = astropy_models.Shift(-7)

    model_with_constraints = astropy_models.Legendre2D(
        x_degree=1, y_degree=1,
        c0_0=1, c0_1=2, c1_0=3,
        fixed={"c1_0": True, "c0_1": True},
        bounds={"c0_0": (-10, 10)}
    )

    model_with_custom_inputs_outputs = astropy_models.Gaussian2D()
    model_with_custom_inputs_outputs.inputs = ("a", "b")
    model_with_custom_inputs_outputs.outputs = ("c",)

    result = [
        # Generic model features
        astropy_models.Shift(10, name="some model name"),
        model_with_bounding_box,
        model_with_user_inverse,
        model_with_constraints,
        model_with_custom_inputs_outputs,

        # astropy.modeling.functional_models
        astropy_models.AiryDisk2D(amplitude=10., x_0=0.5, y_0=1.5),
        astropy_models.Box1D(amplitude=10., x_0=0.5, width=5.),
        astropy_models.Box2D(amplitude=10., x_0=0.5, x_width=5., y_0=1.5, y_width=7.),
        astropy_models.Const1D(amplitude=5.),
        astropy_models.Const2D(amplitude=5.),
        astropy_models.Disk2D(amplitude=10., x_0=0.5, y_0=1.5, R_0=5.),
        astropy_models.Ellipse2D(amplitude=10., x_0=0.5, y_0=1.5, a=2., b=4., theta=0.1),
        astropy_models.Exponential1D(amplitude=10., tau=3.5),
        astropy_models.Gaussian1D(amplitude=10., mean=5., stddev=3.),
        astropy_models.Gaussian2D(amplitude=10., x_mean=5., y_mean=5., x_stddev=3., y_stddev=3.),
        astropy_models.KingProjectedAnalytic1D(amplitude=10., r_core=5., r_tide=2.),
        astropy_models.Linear1D(slope=2.0, intercept=1.5),
        astropy_models.Logarithmic1D(amplitude=10., tau=3.5),
        astropy_models.Lorentz1D(amplitude=10., x_0=0.5, fwhm=2.5),
        astropy_models.Moffat1D(amplitude=10., x_0=0.5, gamma=1.2, alpha=2.5),
        astropy_models.Moffat2D(amplitude=10., x_0=0.5, y_0=1.5, gamma=1.2, alpha=2.5),
        astropy_models.Multiply(3),
        astropy_models.Multiply(10 * u.m),
        astropy_models.Planar2D(slope_x=0.5, slope_y=1.2, intercept=2.5),
        astropy_models.RedshiftScaleFactor(z=2.5),
        astropy_models.RickerWavelet1D(amplitude=10., x_0=0.5, sigma=1.2),
        astropy_models.RickerWavelet2D(amplitude=10., x_0=0.5, y_0=1.5, sigma=1.2),
        astropy_models.Ring2D(amplitude=10., x_0=0.5, y_0=1.5, r_in=5., width=10.),
        astropy_models.Scale(3.4),
        astropy_models.Sersic1D(amplitude=10., r_eff=1., n=4.),
        astropy_models.Sersic2D(amplitude=10., r_eff=1., n=4., x_0=0.5, y_0=1.5, ellip=0.0, theta=0.0),
        astropy_models.Shift(2.),
        astropy_models.Shift(2. * u.deg),
        astropy_models.Scale(3.4 * u.deg),
        astropy_models.Sine1D(amplitude=10., frequency=0.5, phase=1.),
        astropy_models.Trapezoid1D(amplitude=10., x_0=0.5, width=5., slope=1.),
        astropy_models.TrapezoidDisk2D(amplitude=10., x_0=0.5, y_0=1.5, R_0=5., slope=1.),
        astropy_models.Voigt1D(x_0=0.55, amplitude_L=10., fwhm_L=0.5, fwhm_G=0.9),

        # astropy.modeling.mappings
        astropy_models.Identity(2),
        astropy_models.Mapping((0, 1), n_inputs=3),

        # astropy.modeling.math_functions
        astropy.modeling.math_functions.AbsoluteUfunc(),
        astropy.modeling.math_functions.AddUfunc(),
        astropy.modeling.math_functions.ArccosUfunc(),
        astropy.modeling.math_functions.ArccoshUfunc(),
        astropy.modeling.math_functions.ArcsinUfunc(),
        astropy.modeling.math_functions.ArcsinhUfunc(),
        astropy.modeling.math_functions.Arctan2Ufunc(),
        astropy.modeling.math_functions.ArctanUfunc(),
        astropy.modeling.math_functions.ArctanhUfunc(),
        astropy.modeling.math_functions.CbrtUfunc(),
        astropy.modeling.math_functions.CosUfunc(),
        astropy.modeling.math_functions.CoshUfunc(),
        astropy.modeling.math_functions.Deg2radUfunc(),
        astropy.modeling.math_functions.DivideUfunc(),
        astropy.modeling.math_functions.DivmodUfunc(),
        astropy.modeling.math_functions.Exp2Ufunc(),
        astropy.modeling.math_functions.ExpUfunc(),
        astropy.modeling.math_functions.Expm1Ufunc(),
        astropy.modeling.math_functions.FabsUfunc(),
        astropy.modeling.math_functions.Floor_divideUfunc(),
        astropy.modeling.math_functions.FmodUfunc(),
        astropy.modeling.math_functions.HypotUfunc(),
        astropy.modeling.math_functions.Log10Ufunc(),
        astropy.modeling.math_functions.Log1pUfunc(),
        astropy.modeling.math_functions.Log2Ufunc(),
        astropy.modeling.math_functions.LogUfunc(),
        astropy.modeling.math_functions.Logaddexp2Ufunc(),
        astropy.modeling.math_functions.LogaddexpUfunc(),
        astropy.modeling.math_functions.ModUfunc(),
        astropy.modeling.math_functions.MultiplyUfunc(),
        astropy.modeling.math_functions.NegativeUfunc(),
        astropy.modeling.math_functions.PositiveUfunc(),
        astropy.modeling.math_functions.PowerUfunc(),
        astropy.modeling.math_functions.Rad2degUfunc(),
        astropy.modeling.math_functions.ReciprocalUfunc(),
        astropy.modeling.math_functions.RemainderUfunc(),
        astropy.modeling.math_functions.RintUfunc(),
        astropy.modeling.math_functions.SinUfunc(),
        astropy.modeling.math_functions.SinhUfunc(),
        astropy.modeling.math_functions.SqrtUfunc(),
        astropy.modeling.math_functions.SquareUfunc(),
        astropy.modeling.math_functions.SubtractUfunc(),
        astropy.modeling.math_functions.TanUfunc(),
        astropy.modeling.math_functions.TanhUfunc(),
        astropy.modeling.math_functions.True_divideUfunc(),

        # astropy.modeling.physical_models
        astropy_models.BlackBody(scale=10.0, temperature=6000. * u.K),
        astropy_models.Drude1D(amplitude=10.0, x_0=0.5, fwhm=2.5),
        # TODO: NFW

        # astropy.modeling.polynomial
        astropy_models.Chebyshev1D(2, c0=2, c1=3, c2=0.5),
        astropy_models.Chebyshev1D(2, c0=2, c1=3, c2=0.5, domain=(0.0, 1.0), window=(1.5, 2.5)),
        astropy_models.Chebyshev2D(1, 1, c0_0=1, c0_1=2, c1_0=3),
        astropy_models.Chebyshev2D(1, 1, c0_0=1, c0_1=2, c1_0=3,
                                x_domain=(1.0, 2.0),
                                y_domain=(3.0, 4.0),
                                x_window=(5.0, 6.0),
                                y_window=(7.0, 8.0)),
        astropy_models.Hermite1D(2, c0=2, c1=3, c2=0.5),
        astropy_models.Hermite2D(1, 1, c0_0=1, c0_1=2, c1_0=3),
        astropy_models.Legendre1D(2, c0=2, c1=3, c2=0.5),
        astropy_models.Legendre2D(1, 1, c0_0=1, c0_1=2, c1_0=3),
        astropy_models.Polynomial1D(2, c0=1, c1=2, c2=3),
        astropy_models.Polynomial2D(1, c0_0=1, c0_1=2, c1_0=3),

        # astropy.modeling.powerlaws
        astropy_models.BrokenPowerLaw1D(amplitude=10, x_break=0.5, alpha_1=2.0, alpha_2=3.5),
        astropy_models.ExponentialCutoffPowerLaw1D(10, 0.5, 2.0, 7.),
        astropy_models.LogParabola1D(amplitude=10, x_0=0.5, alpha=2., beta=3.,),
        astropy_models.PowerLaw1D(amplitude=10., x_0=0.5, alpha=2.0),
        astropy_models.SmoothlyBrokenPowerLaw1D(amplitude=10., x_break=5.0, alpha_1=2.0, alpha_2=3.0, delta=0.5),

        # astropy.modeling.projections
        astropy_models.AffineTransformation2D(matrix=np.array([[1.0, 2.0], [3.0, 4.0]]), translation=np.array([5.0, 6.0])),
        astropy_models.Pix2Sky_Airy(theta_b=75.8),
        astropy_models.Sky2Pix_Airy(theta_b=75.8),
        astropy_models.Pix2Sky_BonneEqualArea(theta1=44.3),
        astropy_models.Sky2Pix_BonneEqualArea(theta1=44.3),
        astropy_models.Pix2Sky_COBEQuadSphericalCube(),
        astropy_models.Sky2Pix_COBEQuadSphericalCube(),
        astropy_models.Pix2Sky_ConicEqualArea(sigma=89.5, delta=0.5),
        astropy_models.Sky2Pix_ConicEqualArea(sigma=89.5, delta=0.5),
        astropy_models.Pix2Sky_ConicEquidistant(sigma=89.5, delta=0.5),
        astropy_models.Sky2Pix_ConicEquidistant(sigma=89.5, delta=0.5),
        astropy_models.Pix2Sky_ConicOrthomorphic(sigma=89.5, delta=0.5),
        astropy_models.Sky2Pix_ConicOrthomorphic(sigma=89.5, delta=0.5),
        astropy_models.Pix2Sky_ConicPerspective(sigma=89.5, delta=0.5),
        astropy_models.Sky2Pix_ConicPerspective(sigma=89.5, delta=0.5),
        astropy_models.Pix2Sky_CylindricalEqualArea(lam=0.5),
        astropy_models.Sky2Pix_CylindricalEqualArea(lam=0.5),
        astropy_models.Pix2Sky_CylindricalPerspective(mu=1.5, lam=2.4),
        astropy_models.Sky2Pix_CylindricalPerspective(mu=1.5, lam=2.4),
        astropy_models.Pix2Sky_Gnomonic(),
        astropy_models.Sky2Pix_Gnomonic(),
        astropy_models.Pix2Sky_HEALPixPolar(),
        astropy_models.Sky2Pix_HEALPixPolar(),
        astropy_models.Pix2Sky_HEALPix(H=12.0, X=17.0),
        astropy_models.Sky2Pix_HEALPix(H=12.0, X=17.0),
        astropy_models.Pix2Sky_HammerAitoff(),
        astropy_models.Sky2Pix_HammerAitoff(),
        astropy_models.Pix2Sky_Mercator(),
        astropy_models.Sky2Pix_Mercator(),
        astropy_models.Pix2Sky_Molleweide(),
        astropy_models.Sky2Pix_Molleweide(),
        astropy_models.Pix2Sky_Parabolic(),
        astropy_models.Sky2Pix_Parabolic(),
        astropy_models.Pix2Sky_PlateCarree(),
        astropy_models.Sky2Pix_PlateCarree(),
        astropy_models.Pix2Sky_Polyconic(),
        astropy_models.Sky2Pix_Polyconic(),
        astropy_models.Pix2Sky_QuadSphericalCube(),
        astropy_models.Sky2Pix_QuadSphericalCube(),
        astropy_models.Pix2Sky_SansonFlamsteed(),
        astropy_models.Sky2Pix_SansonFlamsteed(),
        astropy_models.Pix2Sky_SlantOrthographic(xi=0.1, eta=0.2),
        astropy_models.Sky2Pix_SlantOrthographic(xi=0.1, eta=0.2),
        astropy_models.Pix2Sky_SlantZenithalPerspective(mu=1.5, phi0=15.0, theta0=80.0),
        astropy_models.Sky2Pix_SlantZenithalPerspective(mu=1.5, phi0=15.0, theta0=80.0),
        astropy_models.Pix2Sky_Stereographic(),
        astropy_models.Sky2Pix_Stereographic(),
        astropy_models.Pix2Sky_TangentialSphericalCube(),
        astropy_models.Sky2Pix_TangentialSphericalCube(),
        astropy_models.Pix2Sky_ZenithalEqualArea(),
        astropy_models.Sky2Pix_ZenithalEqualArea(),
        astropy_models.Pix2Sky_ZenithalEquidistant(),
        astropy_models.Sky2Pix_ZenithalEquidistant(),
        astropy_models.Pix2Sky_ZenithalPerspective(mu=1.5, gamma=15.0),
        astropy_models.Sky2Pix_ZenithalPerspective(mu=1.5, gamma=15.0),

        # astropy.modeling.rotations
        astropy_models.EulerAngleRotation(23, 14, 2.3, axes_order="xzx"),
        astropy_models.RotateCelestial2Native(5.63, -72.5, 180),
        astropy_models.RotateCelestial2Native(5.63 * u.deg, -72.5 * u.deg, 180 * u.deg),
        astropy_models.RotateNative2Celestial(5.63, -72.5, 180),
        astropy_models.RotateNative2Celestial(5.63 * u.deg, -72.5 * u.deg, 180 * u.deg),
        astropy_models.Rotation2D(angle=1.51),
        astropy_models.RotationSequence3D([1.2, 2.3, 3.4, .3], "xyzx"),
        astropy_models.SphericalRotationSequence([1.2, 2.3, 3.4, .3], "xyzy"),

        # astropy.modeling.tabular
        astropy_models.Tabular1D(points=np.arange(0, 5), lookup_table=[1., 10, 2, 45, -3]),
        astropy_models.Tabular1D(points=np.arange(0, 5) * u.pix, lookup_table=[1., 10, 2, 45, -3] * u.nm),
        astropy_models.Tabular2D(
            points=([1, 2, 3], [1, 2, 3]),
            lookup_table=np.arange(0, 9).reshape(3, 3),
            bounds_error=False,
            fill_value=None,
            method="nearest",
        ),
        astropy_models.Tabular2D(
            points=([1, 2, 3], [1, 2, 3]) * u.pix,
            lookup_table=np.arange(0, 9).reshape(3, 3) * u.nm,
            bounds_error=False,
            fill_value=None,
            method="nearest",
        ),
    ]

    if Version(astropy.__version__) >= Version("4.1"):
        result.append(astropy_models.Plummer1D(mass=10.0, r_plum=5.0))

    return result


UNSUPPORTED_MODELS = [
    # FITS-specific and deemed unworthy of ASDF serialization:
    astropy.modeling.polynomial.InverseSIP,
    astropy.modeling.polynomial.SIP,

    # Base classes which should not be directly supported:
    astropy.modeling.core.Model,
    astropy.modeling.math_functions._NPUfuncModel,
    astropy.modeling.polynomial.OrthoPolynomialBase,
    astropy.modeling.polynomial.PolynomialModel,
    astropy.modeling.projections.Conic,
    astropy.modeling.projections.Cylindrical,
    astropy.modeling.projections.HEALPix,
    astropy.modeling.projections.Pix2SkyProjection,
    astropy.modeling.projections.Projection,
    astropy.modeling.projections.PseudoConic,
    astropy.modeling.projections.PseudoCylindrical,
    astropy.modeling.projections.QuadCube,
    astropy.modeling.projections.Sky2PixProjection,
    astropy.modeling.projections.Zenithal,
]


if Version(astropy.__version__) > Version("4.1"):
    # https://github.com/astropy/asdf-astropy/issues/6
    UNSUPPORTED_MODELS += [astropy.modeling.physical_models.NFW,
                           astropy.modeling.models.UnitsMapping,
                          ]

if Version(astropy.__version__) < Version("4.3"):
    UNSUPPORTED_MODELS.append(astropy.modeling.blackbody.BlackBody1D)
else:
    UNSUPPORTED_MODELS.append(astropy.modeling.physical_models.BlackBody)

if Version(astropy.__version__) < Version("4.999.999"):
    UNSUPPORTED_MODELS.extend([
        astropy.modeling.models.MexicanHat1D,
        astropy.modeling.models.MexicanHat2D,
    ])

if Version(astropy.__version__) > Version("4.999.999"):
    # https://github.com/astropy/asdf-astropy/issues/28
    UNSUPPORTED_MODELS.extend([
        astropy.modeling.functional_models.ArcCosine1D,
        astropy.modeling.functional_models.ArcSine1D,
        astropy.modeling.functional_models.ArcTangent1D,
        astropy.modeling.functional_models.Cosine1D,
        astropy.modeling.functional_models.Tangent1D,
        astropy.modeling.spline.Spline1D,
    ])


@pytest.mark.parametrize("model", create_single_models())
def test_single_model(tmpdir, model):
    assert_model_roundtrip(model, tmpdir)


def test_all_models_supported():
    """
    Test that all model classes in astropy have serialization
    support implemented in this package.  If this test fails,
    file an issue on GitHub for each missing model and add
    the model to the UNSUPPORTED_MODELS list above with
    a link to the issue in a comment.
    """
    def _iterate_model_classes():
        for key, value in itertools.chain(astropy_models.__dict__.items(), astropy.modeling.math_functions.__dict__.items()):
            if isinstance(value, type) and issubclass(value, astropy.modeling.core.Model) and value not in UNSUPPORTED_MODELS:
                yield value

    extensions = integration.get_extensions()
    extension_manager = asdf.extension.ExtensionManager(extensions)

    missing = set()
    for model_class in _iterate_model_classes():
        if not extension_manager.handles_type(model_class):
            missing.add(model_class)

    if len(missing) > 0:
        class_names = sorted([f"{m.__module__}.{m.__qualname__}" for m in missing])
        message = "Missing support for the following model classes:\n\n" + "\n".join(class_names)
        assert len(missing) == 0, message


@pytest.mark.xfail(reason="Older tag versions are difficult to test until asdf implements new config features")
def test_legacy_const(tmpdir):
    model = astropy_models.Const1D(amplitude=5.)
    assert_model_roundtrip(model, tmpdir, version="1.3.0")

    model = astropy_models.Const2D(amplitude=5.)
    with pytest.raises(TypeError, match="does not support models with > 1 dimension"):
        assert_model_roundtrip(model, tmpdir, version="1.3.0")


COMPOUND_OPERATORS = [
    "__add__",
    "__sub__",
    "__mul__",
    "__truediv__",
    "__pow__",
    "__or__",
    "__and__",
]


@pytest.mark.parametrize("operator", COMPOUND_OPERATORS)
def test_compound_model(tmpdir, operator):
    left_model = astropy_models.Shift(5)
    right_model = astropy_models.Shift(-1)
    model = getattr(left_model, operator)(right_model)
    result = assert_model_roundtrip(model, tmpdir)
    assert_models_equal(result.left, left_model)
    assert_models_equal(result.right, right_model)
    assert result.op == model.op


def test_fix_inputs(tmpdir):
    model = astropy_models.Gaussian2D(1, 2, 3, 4, 5)
    fixed_model = astropy_models.fix_inputs(model, {"x": 2.5})
    result = assert_model_roundtrip(fixed_model, tmpdir)
    assert_models_equal(result.left, model)
    assert result.right == fixed_model.right
    assert result.op == fixed_model.op


@pytest.mark.skipif('HAS_NO_UNITS_MAPPING')
def test_units_mapping(tmpdir):
    # Basic mapping between units:
    model = astropy_models.UnitsMapping(((u.m, u.dimensionless_unscaled),))
    model.name = "foo"
    result = assert_model_roundtrip(model, tmpdir)
    assert result.mapping == model.mapping

    # Remove units:
    model = astropy_models.UnitsMapping(((u.m, None),))
    result = assert_model_roundtrip(model, tmpdir)
    assert result.mapping == model.mapping

    # Change a model to accept any units:
    model = astropy_models.UnitsMapping(((None, u.m),))
    result = assert_model_roundtrip(model, tmpdir)
    assert result.mapping == model.mapping

    # With equivalencies:
    model = astropy_models.UnitsMapping(((u.m, u.dimensionless_unscaled),),
                                        input_units_equivalencies={"x": u.equivalencies.spectral()})
    result = assert_model_roundtrip(model, tmpdir)
    assert result.mapping == model.mapping

    # Allow dimensionless on all inputs:
    model = astropy_models.UnitsMapping(((u.m, u.dimensionless_unscaled), (u.s, u.Hz)),
                                        input_units_allow_dimensionless=True)
    result = assert_model_roundtrip(model, tmpdir)
    assert result.mapping == model.mapping

    # Allow dimensionless selectively:
    model = astropy_models.UnitsMapping(((u.m, u.dimensionless_unscaled), (u.s, u.Hz)),
                                        input_units_allow_dimensionless={"x0": True, "x1": False})
    result = assert_model_roundtrip(model, tmpdir)
    assert result.mapping == model.mapping


@pytest.mark.parametrize("standard_version", [v for v in asdf.versioning.supported_versions if v >= "1.4.0"])
@pytest.mark.parametrize("model", [
    astropy_models.Polynomial1D(1, c0=5, c1=17),
    astropy_models.Polynomial1D(1, c0=5, c1=17, domain=[-5, 4], window=[-2, 3]),
    astropy_models.Chebyshev1D(2, c0=2, c1=3, c2=0.5, domain=[-2, 2]),
    astropy_models.Chebyshev1D(2, c0=2, c1=3, c2=0.5,
                               domain=[-2, 2], window=[-0.5, 0.5]),
])
def test_1d_polynomial_with_asdf_standard_version(tmpdir, standard_version, model):
    result = assert_model_roundtrip(model, tmpdir, version=standard_version)
    assert result.domain == model.domain
    assert result.window == model.window


@pytest.mark.parametrize("standard_version", [v for v in asdf.versioning.supported_versions if v >= "1.4.0"])
@pytest.mark.parametrize("model", [
    astropy_models.Polynomial2D(2, c0_0=3, c1_0=5, c0_1=7),
    astropy_models.Polynomial2D(
        2, c0_0=3, c1_0=5, c0_1=7, x_domain=[-2, 2], y_domain=[-4, 4],
        x_window=[-6, 6], y_window=[-8, 8]
    ),
    astropy_models.Chebyshev2D(1, 1, c0_0=1, c0_1=2, c1_0=3,
                               x_domain=[-2, 2], y_domain=[-2, 2]),
    astropy_models.Chebyshev2D(1, 1, c0_0=1, c0_1=2, c1_0=3,
                               x_domain=[-2, 2], y_domain=[-2, 2],
                               x_window=[-0.5, 0.5], y_window=[-0.1, 0.5]),
])
def test_2d_polynomial_with_asdf_standard_version(tmpdir, standard_version, model):
    result = assert_model_roundtrip(model, tmpdir, version=standard_version)
    assert result.x_domain == model.x_domain
    assert result.y_domain == model.y_domain
    assert result.x_window == model.x_window
    assert result.y_window == model.y_window


@pytest.mark.skip('ASTROPY_LT_41')
def test_deserialize_compound_user_inverse(tmpdir):
    """
    Confirm that we are able to correctly reconstruct a
    compound model with a user inverse set on one of its
    component models.

    Due to code in TransformConverter that facilitates circular
    inverses, the user inverse of the component model is
    not available at the time that the CompoundModel is
    constructed.
    """

    yaml = """
model: !transform/concatenate-1.2.0
  forward:
  - !transform/shift-1.2.0
    inverse: !transform/shift-1.2.0 {offset: 5.0}
    offset: -10.0
  - !transform/shift-1.2.0 {offset: -20.0}
  """
    buff = yaml_to_asdf(yaml)
    with asdf.open(buff) as af:
        model = af["model"]
        assert model.has_inverse()
        assert model.inverse(-5, -20) == (0, 0)



from asdf.extension import ManifestExtension

from .converters.transform.compound import CompoundConverter
from .converters.transform.core import SimpleTransformConverter
from .converters.transform.functional_models import ConstantConverter
from .converters.transform.mappings import (
    IdentityConverter,
    RemapAxesConverter,
    UnitsMappingConverter,
)
from .converters.transform.math_functions import MathFunctionsConverter
from .converters.transform.polynomial import PolynomialConverter, OrthoPolynomialConverter
from .converters.transform.projections import ProjectionConverter
from .converters.transform.rotations import Rotate3DConverter, RotationSequenceConverter
from .converters.transform.tabular import TabularConverter

from .converters.unit.equivalency import EquivalencyConverter
from .converters.unit.quantity import QuantityConverter
from .converters.unit.unit import UnitConverter

from .converters.coordinates.angle import AngleConverter, LatitudeConverter, LongitudeConverter
from .converters.coordinates.earth_location import EarthLocationConverter
from .converters.coordinates.frame import FrameConverter, LegacyICRSConverter
from .converters.coordinates.representation import RepresentationConverter
from .converters.coordinates.sky_coord import SkyCoordConverter
from .converters.coordinates.spectral_coord import SpectralCoordConverter

from .converters.time.time import TimeConverter
from .converters.time.time_delta import TimeDeltaConverter

from .converters.table.table import ColumnConverter, AsdfTableConverter, AstropyTableConverter

from .converters.fits.fits import AstropyFitsConverter, AsdfFitsConverter


TRANSFORM_CONVERTERS = [
    # astropy.modeling.core
    CompoundConverter(),

    # astropy.modeling.functional_models
    SimpleTransformConverter(
        ["tag:stsci.edu:asdf/transform/airy_disk2d-*"],
        "astropy.modeling.functional_models.AiryDisk2D",
    ),
    SimpleTransformConverter(
        ["tag:stsci.edu:asdf/transform/box1d-*"],
        "astropy.modeling.functional_models.Box1D",
    ),
    SimpleTransformConverter(
        ["tag:stsci.edu:asdf/transform/box2d-*"],
        "astropy.modeling.functional_models.Box2D",
    ),
    ConstantConverter(),
    SimpleTransformConverter(
        ["tag:stsci.edu:asdf/transform/disk2d-*"],
        "astropy.modeling.functional_models.Disk2D",
    ),
    SimpleTransformConverter(
        ["tag:stsci.edu:asdf/transform/ellipse2d-*"],
        "astropy.modeling.functional_models.Ellipse2D",
    ),
    SimpleTransformConverter(
        ["tag:stsci.edu:asdf/transform/exponential1d-*"],
        "astropy.modeling.functional_models.Exponential1D",
    ),
    SimpleTransformConverter(
        ["tag:stsci.edu:asdf/transform/gaussian1d-*"],
        "astropy.modeling.functional_models.Gaussian1D",
    ),
    SimpleTransformConverter(
        ["tag:stsci.edu:asdf/transform/gaussian2d-*"],
        "astropy.modeling.functional_models.Gaussian2D",
    ),
    SimpleTransformConverter(
        ["tag:stsci.edu:asdf/transform/king_projected_analytic1d-*"],
        "astropy.modeling.functional_models.KingProjectedAnalytic1D",
    ),
    SimpleTransformConverter(
        ["tag:stsci.edu:asdf/transform/linear1d-*"],
        "astropy.modeling.functional_models.Linear1D",
    ),
    SimpleTransformConverter(
        ["tag:stsci.edu:asdf/transform/logarithmic1d-*"],
        "astropy.modeling.functional_models.Logarithmic1D",
    ),
    SimpleTransformConverter(
        ["tag:stsci.edu:asdf/transform/lorentz1d-*"],
        "astropy.modeling.functional_models.Lorentz1D",
    ),
    SimpleTransformConverter(
        ["tag:stsci.edu:asdf/transform/moffat1d-*"],
        "astropy.modeling.functional_models.Moffat1D",
    ),
    SimpleTransformConverter(
        ["tag:stsci.edu:asdf/transform/moffat2d-*"],
        "astropy.modeling.functional_models.Moffat2D",
    ),
    SimpleTransformConverter(
        ["tag:stsci.edu:asdf/transform/multiplyscale-*"],
        "astropy.modeling.functional_models.Multiply",
    ),
    SimpleTransformConverter(
        ["tag:stsci.edu:asdf/transform/planar2d-*"],
        "astropy.modeling.functional_models.Planar2D",
    ),
    SimpleTransformConverter(
        ["tag:stsci.edu:asdf/transform/redshift_scale_factor-*"],
        "astropy.modeling.functional_models.RedshiftScaleFactor",
    ),
    SimpleTransformConverter(
        ["tag:stsci.edu:asdf/transform/ricker_wavelet1d-*"],
        "astropy.modeling.functional_models.RickerWavelet1D",
    ),
    SimpleTransformConverter(
        ["tag:stsci.edu:asdf/transform/ricker_wavelet2d-*"],
        "astropy.modeling.functional_models.RickerWavelet2D",
    ),
    SimpleTransformConverter(
        ["tag:stsci.edu:asdf/transform/ring2d-*"],
        "astropy.modeling.functional_models.Ring2D",
    ),
    SimpleTransformConverter(
        ["tag:stsci.edu:asdf/transform/scale-*"],
        "astropy.modeling.functional_models.Scale",
    ),
    SimpleTransformConverter(
        ["tag:stsci.edu:asdf/transform/sersic1d-*"],
        "astropy.modeling.functional_models.Sersic1D",
    ),
    SimpleTransformConverter(
        ["tag:stsci.edu:asdf/transform/sersic2d-*"],
        "astropy.modeling.functional_models.Sersic2D",
    ),
    SimpleTransformConverter(
        ["tag:stsci.edu:asdf/transform/shift-*"],
        "astropy.modeling.functional_models.Shift",
    ),
    SimpleTransformConverter(
        ["tag:stsci.edu:asdf/transform/sine1d-*"],
        "astropy.modeling.functional_models.Sine1D",
    ),
    SimpleTransformConverter(
        ["tag:stsci.edu:asdf/transform/trapezoid1d-*"],
        "astropy.modeling.functional_models.Trapezoid1D",
    ),
    SimpleTransformConverter(
        ["tag:stsci.edu:asdf/transform/trapezoid_disk2d-*"],
        "astropy.modeling.functional_models.TrapezoidDisk2D",
    ),
    SimpleTransformConverter(
        ["tag:stsci.edu:asdf/transform/voigt1d-*"],
        "astropy.modeling.functional_models.Voigt1D",
    ),

    # astropy.modeling.mappings
    IdentityConverter(),
    RemapAxesConverter(),
    # UnitsMapping is not represented here because
    # it is an astropy-specific transform and not
    # included in the ASDF transform extension.

    # astropy.modeling.math_functions
    MathFunctionsConverter(),

    # astropy.modeling.physical_models
    SimpleTransformConverter(
        ["tag:stsci.edu:asdf/transform/blackbody-*"],
        "astropy.modeling.physical_models.BlackBody",
    ),
    SimpleTransformConverter(
        ["tag:stsci.edu:asdf/transform/drude1d-*"],
        "astropy.modeling.physical_models.Drude1D",
    ),
    # TODO: Implement NFW
    SimpleTransformConverter(
        ["tag:stsci.edu:asdf/transform/plummer1d-*"],
        "astropy.modeling.physical_models.Plummer1D",
    ),

    # astropy.modeling.polynomial
    PolynomialConverter(),
    OrthoPolynomialConverter(),
    # SIP and InverseSIP are deliberately excluded because they
    # are FITS-specific and can be easily represented by a
    # simple combination of existing models.

    # astropy.modeling.powerlaws
    SimpleTransformConverter(
        ["tag:stsci.edu:asdf/transform/broken_power_law1d-*"],
        "astropy.modeling.powerlaws.BrokenPowerLaw1D",
    ),
    SimpleTransformConverter(
        ["tag:stsci.edu:asdf/transform/exponential_cutoff_power_law1d-*"],
        "astropy.modeling.powerlaws.ExponentialCutoffPowerLaw1D",
    ),
    SimpleTransformConverter(
        ["tag:stsci.edu:asdf/transform/log_parabola1d-*"],
        "astropy.modeling.powerlaws.LogParabola1D",
    ),
    SimpleTransformConverter(
        ["tag:stsci.edu:asdf/transform/power_law1d-*"],
        "astropy.modeling.powerlaws.PowerLaw1D",
    ),
    SimpleTransformConverter(
        ["tag:stsci.edu:asdf/transform/smoothly_broken_power_law1d-*"],
        "astropy.modeling.powerlaws.SmoothlyBrokenPowerLaw1D",
    ),

    # astropy.modeling.projections
    SimpleTransformConverter(
        ["tag:stsci.edu:asdf/transform/affine-*"],
        "astropy.modeling.projections.AffineTransformation2D",
    ),
    ProjectionConverter(
        ["tag:stsci.edu:asdf/transform/airy-*"],
        "astropy.modeling.projections.Pix2Sky_Airy",
        "astropy.modeling.projections.Sky2Pix_Airy",
    ),
    ProjectionConverter(
        ["tag:stsci.edu:asdf/transform/bonne_equal_area-*"],
        "astropy.modeling.projections.Pix2Sky_BonneEqualArea",
        "astropy.modeling.projections.Sky2Pix_BonneEqualArea",
    ),
    ProjectionConverter(
        ["tag:stsci.edu:asdf/transform/cobe_quad_spherical_cube-*"],
        "astropy.modeling.projections.Pix2Sky_COBEQuadSphericalCube",
        "astropy.modeling.projections.Sky2Pix_COBEQuadSphericalCube",
    ),
    ProjectionConverter(
        ["tag:stsci.edu:asdf/transform/conic_equal_area-*"],
        "astropy.modeling.projections.Pix2Sky_ConicEqualArea",
        "astropy.modeling.projections.Sky2Pix_ConicEqualArea",
    ),
    ProjectionConverter(
        ["tag:stsci.edu:asdf/transform/conic_equidistant-*"],
        "astropy.modeling.projections.Pix2Sky_ConicEquidistant",
        "astropy.modeling.projections.Sky2Pix_ConicEquidistant",
    ),
    ProjectionConverter(
        ["tag:stsci.edu:asdf/transform/conic_orthomorphic-*"],
        "astropy.modeling.projections.Pix2Sky_ConicOrthomorphic",
        "astropy.modeling.projections.Sky2Pix_ConicOrthomorphic",
    ),
    ProjectionConverter(
        ["tag:stsci.edu:asdf/transform/conic_perspective-*"],
        "astropy.modeling.projections.Pix2Sky_ConicPerspective",
        "astropy.modeling.projections.Sky2Pix_ConicPerspective",
    ),
    ProjectionConverter(
        ["tag:stsci.edu:asdf/transform/cylindrical_equal_area-*"],
        "astropy.modeling.projections.Pix2Sky_CylindricalEqualArea",
        "astropy.modeling.projections.Sky2Pix_CylindricalEqualArea",
    ),
    ProjectionConverter(
        ["tag:stsci.edu:asdf/transform/cylindrical_perspective-*"],
        "astropy.modeling.projections.Pix2Sky_CylindricalPerspective",
        "astropy.modeling.projections.Sky2Pix_CylindricalPerspective",
    ),
    ProjectionConverter(
        ["tag:stsci.edu:asdf/transform/gnomonic-*"],
        "astropy.modeling.projections.Pix2Sky_Gnomonic",
        "astropy.modeling.projections.Sky2Pix_Gnomonic",
    ),
    ProjectionConverter(
        ["tag:stsci.edu:asdf/transform/hammer_aitoff-*"],
        "astropy.modeling.projections.Pix2Sky_HammerAitoff",
        "astropy.modeling.projections.Sky2Pix_HammerAitoff",
    ),
    ProjectionConverter(
        ["tag:stsci.edu:asdf/transform/healpix-*"],
        "astropy.modeling.projections.Pix2Sky_HEALPix",
        "astropy.modeling.projections.Sky2Pix_HEALPix",
    ),
    ProjectionConverter(
        ["tag:stsci.edu:asdf/transform/healpix_polar-*"],
        "astropy.modeling.projections.Pix2Sky_HEALPixPolar",
        "astropy.modeling.projections.Sky2Pix_HEALPixPolar",
    ),
    ProjectionConverter(
        ["tag:stsci.edu:asdf/transform/mercator-*"],
        "astropy.modeling.projections.Pix2Sky_Mercator",
        "astropy.modeling.projections.Sky2Pix_Mercator",
    ),
    ProjectionConverter(
        ["tag:stsci.edu:asdf/transform/molleweide-*"],
        "astropy.modeling.projections.Pix2Sky_Molleweide",
        "astropy.modeling.projections.Sky2Pix_Molleweide",
    ),
    ProjectionConverter(
        ["tag:stsci.edu:asdf/transform/parabolic-*"],
        "astropy.modeling.projections.Pix2Sky_Parabolic",
        "astropy.modeling.projections.Sky2Pix_Parabolic",
    ),
    ProjectionConverter(
        ["tag:stsci.edu:asdf/transform/plate_carree-*"],
        "astropy.modeling.projections.Pix2Sky_PlateCarree",
        "astropy.modeling.projections.Sky2Pix_PlateCarree",
    ),
    ProjectionConverter(
        ["tag:stsci.edu:asdf/transform/polyconic-*"],
        "astropy.modeling.projections.Pix2Sky_Polyconic",
        "astropy.modeling.projections.Sky2Pix_Polyconic",
    ),
    ProjectionConverter(
        ["tag:stsci.edu:asdf/transform/quad_spherical_cube-*"],
        "astropy.modeling.projections.Pix2Sky_QuadSphericalCube",
        "astropy.modeling.projections.Sky2Pix_QuadSphericalCube",
    ),
    ProjectionConverter(
        ["tag:stsci.edu:asdf/transform/sanson_flamsteed-*"],
        "astropy.modeling.projections.Pix2Sky_SansonFlamsteed",
        "astropy.modeling.projections.Sky2Pix_SansonFlamsteed",
    ),
    ProjectionConverter(
        ["tag:stsci.edu:asdf/transform/slant_orthographic-*"],
        "astropy.modeling.projections.Pix2Sky_SlantOrthographic",
        "astropy.modeling.projections.Sky2Pix_SlantOrthographic",
    ),
    ProjectionConverter(
        ["tag:stsci.edu:asdf/transform/slant_zenithal_perspective-*"],
        "astropy.modeling.projections.Pix2Sky_SlantZenithalPerspective",
        "astropy.modeling.projections.Sky2Pix_SlantZenithalPerspective",
    ),
    ProjectionConverter(
        ["tag:stsci.edu:asdf/transform/stereographic-*"],
        "astropy.modeling.projections.Pix2Sky_Stereographic",
        "astropy.modeling.projections.Sky2Pix_Stereographic",
    ),
    ProjectionConverter(
        ["tag:stsci.edu:asdf/transform/tangential_spherical_cube-*"],
        "astropy.modeling.projections.Pix2Sky_TangentialSphericalCube",
        "astropy.modeling.projections.Sky2Pix_TangentialSphericalCube",
    ),
    ProjectionConverter(
        ["tag:stsci.edu:asdf/transform/zenithal_equal_area-*"],
        "astropy.modeling.projections.Pix2Sky_ZenithalEqualArea",
        "astropy.modeling.projections.Sky2Pix_ZenithalEqualArea",
    ),
    ProjectionConverter(
        ["tag:stsci.edu:asdf/transform/zenithal_equidistant-*"],
        "astropy.modeling.projections.Pix2Sky_ZenithalEquidistant",
        "astropy.modeling.projections.Sky2Pix_ZenithalEquidistant",
    ),
    ProjectionConverter(
        ["tag:stsci.edu:asdf/transform/zenithal_perspective-*"],
        "astropy.modeling.projections.Pix2Sky_ZenithalPerspective",
        "astropy.modeling.projections.Sky2Pix_ZenithalPerspective",
    ),

    # astropy.modeling.rotations
    SimpleTransformConverter(
        ["tag:stsci.edu:asdf/transform/rotate2d-*"],
        "astropy.modeling.rotations.Rotation2D",
    ),
    Rotate3DConverter(),
    RotationSequenceConverter(),

    # astropy.modeling.tabular
    TabularConverter(),
]


# The order here is important; asdf will prefer to use extensions
# that occur earlier in the list.
TRANSFORM_MANIFEST_URIS = [
    "asdf://asdf-format.org/transform/manifests/transform-1.5.0",
    "asdf://asdf-format.org/transform/manifests/transform-1.4.0",
    "asdf://asdf-format.org/transform/manifests/transform-1.3.0",
    "asdf://asdf-format.org/transform/manifests/transform-1.2.0",
    "asdf://asdf-format.org/transform/manifests/transform-1.1.0",
    "asdf://asdf-format.org/transform/manifests/transform-1.0.0",
]


TRANSFORM_EXTENSIONS = [
    ManifestExtension.from_uri(
        uri,
        # This prevents a warning about a missing extension when opening
        # files written by older versions of the asdf library:
        legacy_class_names=["astropy.io.misc.asdf.extension.AstropyAsdfExtension"],
        converters=TRANSFORM_CONVERTERS,
    )
    for uri in TRANSFORM_MANIFEST_URIS
]


COORDINATES_CONVERTERS = [
    FrameConverter(
        "tag:astropy.org:astropy/coordinates/frames/baseframe-*",
        "astropy.coordinates.baseframe.BaseCoordinateFrame",
    ),
    FrameConverter(
        "tag:astropy.org:astropy/coordinates/frames/cirs-*",
        "astropy.coordinates.builtin_frames.cirs.CIRS",
    ),
    FrameConverter(
        "tag:astropy.org:astropy/coordinates/frames/fk4-*",
        "astropy.coordinates.builtin_frames.fk4.FK4",
    ),
    FrameConverter(
        "tag:astropy.org:astropy/coordinates/frames/fk4noeterms-*",
        "astropy.coordinates.builtin_frames.fk4.FK4NoETerms",
    ),
    FrameConverter(
        "tag:astropy.org:astropy/coordinates/frames/fk5-*",
        "astropy.coordinates.builtin_frames.fk5.FK5",
    ),
    FrameConverter(
        "tag:astropy.org:astropy/coordinates/frames/galactic-*",
        "astropy.coordinates.builtin_frames.galactic.Galactic",
    ),
    FrameConverter(
        "tag:astropy.org:astropy/coordinates/frames/galactocentric-*",
        "astropy.coordinates.builtin_frames.galactocentric.Galactocentric",
    ),
    FrameConverter(
        "tag:astropy.org:astropy/coordinates/frames/gcrs-*",
        "astropy.coordinates.builtin_frames.gcrs.GCRS",
    ),
    FrameConverter(
        "tag:astropy.org:astropy/coordinates/frames/icrs-1.1.0",
        "astropy.coordinates.builtin_frames.icrs.ICRS",
    ),
    FrameConverter(
        "tag:astropy.org:astropy/coordinates/frames/itrs-*",
        "astropy.coordinates.builtin_frames.itrs.ITRS",
    ),
    FrameConverter(
        "tag:astropy.org:astropy/coordinates/frames/precessedgeocentric-*",
        "astropy.coordinates.builtin_frames.gcrs.PrecessedGeocentric",
    ),
    LegacyICRSConverter(),
    AngleConverter(),
    LatitudeConverter(),
    LongitudeConverter(),
    EarthLocationConverter(),
    RepresentationConverter(),
    SkyCoordConverter(),
    SpectralCoordConverter(),
]


ASTROPY_CONVERTERS = [
    EquivalencyConverter(),
    UnitsMappingConverter(),
    TimeDeltaConverter(),
    AstropyTableConverter(),
    AstropyFitsConverter(),
]


COORDINATES_EXTENSION = ManifestExtension.from_uri(
    "asdf://asdf-format.org/astronomy/coordinates/manifests/coordinates-1.0.0",
    converters=COORDINATES_CONVERTERS,
)


ASTROPY_EXTENSION = ManifestExtension.from_uri(
    "asdf://astropy.org/astropy/manifests/astropy-1.0.0",
    # This prevents a warning about a missing extension when opening
    # files written by older versions of astropy:
    legacy_class_names=["astropy.io.misc.asdf.extension.AstropyExtension"],
    converters=ASTROPY_CONVERTERS,
)

# These tags are part of the ASDF Standard,
# but we want to override serialization here so that users can
# work with nice astropy objects for those entities.

CORE_CONVERTERS = [
    QuantityConverter(),
    TimeConverter(),
    UnitConverter(),
    ColumnConverter(),
    AsdfTableConverter(),
    AsdfFitsConverter(),
]


CORE_MANIFEST_URIS = [
    "asdf://asdf-format.org/core/manifests/core-1.0.0",
    "asdf://asdf-format.org/core/manifests/core-1.1.0",
    "asdf://asdf-format.org/core/manifests/core-1.2.0",
    "asdf://asdf-format.org/core/manifests/core-1.3.0",
    "asdf://asdf-format.org/core/manifests/core-1.4.0",
    "asdf://asdf-format.org/core/manifests/core-1.5.0",
    "asdf://asdf-format.org/core/manifests/core-1.6.0",
]


CORE_EXTENSIONS = [ManifestExtension.from_uri(u, converters=CORE_CONVERTERS) for u in CORE_MANIFEST_URIS]



import pytest

import asdf
from asdf.tests import helpers
from astropy import units
from astropy.units import Quantity
import numpy as np
from numpy.testing import assert_array_equal


def create_quantities():
    return [
        # Scalar:
        Quantity(2.71828, units.kpc),
        # Single element array:
        Quantity([3.14159], units.kg),
        # Multiple element array:
        Quantity([x * 2.3081 for x in range(10)], units.ampere),
        # Multiple dimension array:
        Quantity(np.arange(100, dtype=np.float64).reshape(5, 20), units.km),
    ]


@pytest.mark.parametrize("quantity", create_quantities())
def test_serialization(quantity, tmp_path):
    file_path = tmp_path / "test.asdf"
    with asdf.AsdfFile() as af:
        af["quantity"] = quantity
        af.write_to(file_path)

    with asdf.open(file_path) as af:
        assert (af["quantity"] == quantity).all()


def test_read_untagged_unit():
    yaml = """
quantity: !unit/quantity-1.1.0
  value: 2.71828
  unit: kpc
    """
    buff = helpers.yaml_to_asdf(yaml)
    with asdf.open(buff) as af:
        assert af["quantity"].value == 2.71828
        assert af["quantity"].unit.is_equivalent(units.kpc)


def test_read_tagged_unit():
    yaml = """
quantity: !unit/quantity-1.1.0
  value: 2.71828
  unit: !unit/unit-1.0.0 kpc
    """
    buff = helpers.yaml_to_asdf(yaml)
    with asdf.open(buff) as af:
        assert af["quantity"].value == 2.71828
        assert af["quantity"].unit.is_equivalent(units.kpc)


def test_read_array_value():
    yaml = """
quantity: !unit/quantity-1.1.0
  value: !core/ndarray-1.0.0 [1.0, 2.0, 3.0, 4.0]
  unit: km
    """
    buff = helpers.yaml_to_asdf(yaml)
    with asdf.open(buff) as af:
        assert_array_equal(af["quantity"].value, np.array([1.0, 2.0, 3.0, 4.0]))
        assert af["quantity"].unit.is_equivalent(units.km)



from asdf.extension import Converter
from asdf.tags.core.ndarray import NDArrayType


class ColumnConverter(Converter):
    tags = ["tag:stsci.edu:asdf/core/column-*"]

    types = [
        "astropy.table.column.Column",
        "astropy.table.column.MaskedColumn",
    ]

    def to_yaml_tree(self, obj, tag, ctx):
        node = {
            "data": obj.data,
            "name": obj.name
        }

        if obj.description:
            node["description"] = obj.description

        if obj.unit:
            node["unit"] = obj.unit

        if obj.meta:
            node["meta"] = obj.meta

        return node

    def from_yaml_tree(self, node, tag, ctx):
        from astropy.table import Column, MaskedColumn
        from numpy.ma.core import MaskedArray

        data = node["data"]
        if isinstance(data, NDArrayType):
            # TODO: Why doesn't NDArrayType work?  This needs some research
            # and documentation.
            data = data._make_array()

        if isinstance(data, MaskedArray):
            column_class = MaskedColumn
        else:
            column_class = Column

        return column_class(
            data=data,
            name=node["name"],
            description=node.get("description"),
            unit=node.get("unit"),
            meta=node.get("meta"),
        )


class AsdfTableConverter(Converter):
    tags = ["tag:stsci.edu:asdf/core/table-*"]

    types = []

    def to_yaml_tree(self, obj, tag, ctx):
        raise NotImplementedError("astropy does not support writing astropy.table.Table with the ASDF table-1.0.0 tag")

    def from_yaml_tree(self, node, tag, ctx):
        from astropy.table import Table

        return Table(node["columns"], meta=node["meta"])


class AstropyTableConverter(Converter):
    tags = ["tag:astropy.org:astropy/table/table-*"]

    types = [
        "astropy.table.table.Table",
        "astropy.table.table.QTable",
    ]

    def to_yaml_tree(self, obj, tag, ctx):
        from astropy.table import QTable

        node = {
            "columns": [obj[name] for name in obj.colnames],
            "colnames": obj.colnames,
            "qtable": isinstance(obj, QTable)
        }

        if obj.meta:
            node["meta"] = obj.meta

        return node

    def from_yaml_tree(self, node, tag, ctx):
        from astropy.table import Table, QTable

        if node.get("qtable", False):
            table = QTable(meta=node.get("meta"))
        else:
            table = Table(meta=node.get("meta"))

        for name, column in zip(node["colnames"], node["columns"]):
            table[name] = column

        return table




import pytest

import asdf

from asdf_astropy.testing.helpers import assert_time_delta_equal

from astropy import units as u
from astropy.time import Time, TimeDelta


def create_time_deltas():
    result = [
        TimeDelta([1, 2] * u.day),
    ]

    for format in TimeDelta.FORMATS.keys():
        result.append(TimeDelta(Time.now() - Time.now(), format=format))

    for scale in list(TimeDelta.SCALES) + [None]:
        result.append(TimeDelta(0.125, scale=scale))

    return result


@pytest.mark.parametrize("time_delta", create_time_deltas())
def test_serialization(time_delta, tmp_path):
    file_path = tmp_path / "test.asdf"
    with asdf.AsdfFile() as af:
        af["time_delta"] = time_delta
        af.write_to(file_path)

    with asdf.open(file_path) as af:
        assert_time_delta_equal(af["time_delta"], time_delta)



from asdf.extension import Converter
from asdf.tags.core.ndarray import NDArrayType


def _card_to_node(card):
    from astropy.io.fits import Undefined

    if isinstance(card.value, Undefined):
        value = ""
    else:
        value = card.value

    if card.comment:
        return [card.keyword, value, card.comment]
    elif value:
        return [card.keyword, value]
    elif card.keyword:
        return [card.keyword]
    else:
        return []


class FitsConverter(Converter):
    def to_yaml_tree(self, obj, tag, ctx):
        from astropy.table import Table

        node = []
        for hdu in obj:
            header_node =  [_card_to_node(c) for c in hdu.header.cards]
            hdu_node = {"header": header_node}

            if hdu.data is not None:
                if hdu.data.dtype.names is not None:
                    hdu_node["data"] = Table(hdu.data)
                else:
                    hdu_node["data"] = hdu.data

            node.append(hdu_node)

        return node

    def from_yaml_tree(self, node, tag, ctx):
        from astropy.io import fits

        hdus = []
        first = True
        for hdu_node in node:
            header = fits.Header([fits.Card(*x) for x in hdu_node["header"]])

            data = hdu_node.get("data")
            if isinstance(data, NDArrayType):
                # TODO: Why doesn't NDArrayType work?  This needs some research
                # and documentation.
                data = data._make_array()

            if first:
                hdu = fits.PrimaryHDU(data=data, header=header)
                first = False
            elif data.dtype.names is not None:
                hdu = fits.BinTableHDU(data=data, header=header)
            else:
                hdu = fits.ImageHDU(data=data, header=header)

            hdus.append(hdu)

        return fits.HDUList(hdus)


class AsdfFitsConverter(FitsConverter):
    tags = ["tag:stsci.edu:asdf/fits/fits-*"]
    types = []


class AstropyFitsConverter(FitsConverter):
    tags = ["tag:astropy.org:astropy/fits/fits-*"]
    types = ["astropy.io.fits.hdu.hdulist.HDUList"]



from packaging.version import parse as parse_version


def parse_tag_version(tag):
    """
    Parse the version portion of the tag into a comparable
    version object.

    Parameters
    ----------
    tag : str

    Returns
    -------
    packaging.version.Version
    """
    return parse_version(tag[tag.rfind("-") + 1:])


def get_tag_name(tag):
    """
    Extract the name portion of a tag URI.

    Parameters
    ----------
    tag : str

    Returns
    -------
    str
    """
    return tag[tag.rfind("/") + 1:tag.rfind("-")]



from astropy.tests.helper import assert_quantity_allclose


def assert_representation_equal(a, b):
    __tracebackhide__ = True

    assert type(a) is type(b)
    assert a.components == b.components
    for component in a.components:
        assert u.allclose(getattr(a, component), getattr(b, component))


def assert_sky_coord_equal(a, b):
    __tracebackhide__ = True

    assert a.is_equivalent_frame(b)
    assert a.representation_type is b.representation_type
    assert a.shape == b.shape

    assert_representation_equal(a.data, b.data)


def assert_frame_equal(a, b):
    __tracebackhide__ = True

    assert type(a) is type(b)

    if a is None:
        return

    assert_representation_equal(a.data, b.data)


def assert_spectral_coord_equal(a, b):
    __tracebackhide__ = True

    assert type(a) is type(b)
    assert_quantity_allclose(a.quantity, b.quantity)
    assert_frame_equal(a.observer, b.observer)
    assert_frame_equal(a.target, b.target)



import warnings

import pytest

# No spectral_coordinate module in astropy 4.0.x
pytest.importorskip("astropy.coordinates.spectral_coordinate")

import asdf
from astropy import units as u
from astropy.coordinates import SpectralCoord, ICRS, Galactic
from astropy.coordinates.spectral_coordinate import NoVelocityWarning
from astropy.tests.helper import assert_quantity_allclose

from asdf_astropy.testing.helpers import assert_frame_equal


def create_spectral_coords():
    result = [
        # Scalar
        SpectralCoord(565 * u.nm),
        # Vector
        SpectralCoord([100, 200, 300] * u.GHz),
    ]

    with warnings.catch_warnings():
        warnings.simplefilter("ignore", NoVelocityWarning)
        result.append(
            # With observer and target
            SpectralCoord(
                10 * u.GHz,
                observer=ICRS(1 * u.km, 2 * u.km, 3 * u.km, representation_type="cartesian"),
                target=Galactic(10 * u.deg, 20 * u.deg, distance=30 * u.pc)
            )
        )

    return result


@pytest.mark.filterwarnings("ignore::astropy.coordinates.spectral_coordinate.NoVelocityWarning")
@pytest.mark.parametrize("coord", create_spectral_coords())
def test_serialization(coord, tmp_path):
    file_path = tmp_path / "test.asdf"

    with asdf.AsdfFile() as af:
        af["coord"] = coord
        af.write_to(file_path)

    with asdf.open(file_path) as af:
        assert type(af["coord"]) is type(coord)
        assert_quantity_allclose(af["coord"].quantity, coord.quantity)
        assert_frame_equal(af["coord"].observer, coord.observer)
        assert_frame_equal(af["coord"].target, coord.target)



from asdf.extension import Converter


class EquivalencyConverter(Converter):
    tags = ["tag:astropy.org:astropy/units/equivalency-*"]

    types = ["astropy.units.equivalencies.Equivalency"]

    def to_yaml_tree(self, obj, tag, ctx):
        return [
            {'name': name, 'kwargs_names': list(kw.keys()), 'kwargs_values': list(kw.values())}
            for name, kw in zip(obj.name, obj.kwargs)
        ]

    def from_yaml_tree(self, node, tag, ctx):
        from astropy.units import equivalencies

        components = []
        for equivalency_node in node:
            equivalency_method = getattr(equivalencies, equivalency_node["name"])
            kwargs = dict(zip(equivalency_node["kwargs_names"], equivalency_node["kwargs_values"]))
            components.append(equivalency_method(**kwargs))

        # The Equivalency class is a UserList that overrides __add__ to
        # provide special behavior when combined with another Equivalency.
        # We're using sum here to add each subsequent Equivalency to the
        # first so we end up with a single correctly combined Equivalency
        # object at the end.
        return sum(components[1:], components[0])



from .core import TransformConverterBase


from asdf.extension import Converter


class IdentityConverter(TransformConverterBase):
    """
    ASDF support for serializing the Identity model.
    """
    tags = ["tag:stsci.edu:asdf/transform/identity-*"]

    types = ["astropy.modeling.mappings.Identity"]

    def to_yaml_tree_transform(self, model, tag, ctx):
        node = {}
        if model.n_inputs != 1:
            node['n_dims'] = model.n_inputs
        return node

    def from_yaml_tree_transform(self, node, tag, ctx):
        from astropy.modeling.mappings import Identity

        return Identity(node.get('n_dims', 1))


class RemapAxesConverter(TransformConverterBase):
    """
    ASDF support for serializing the Mapping model
    """
    tags = ["tag:stsci.edu:asdf/transform/remap_axes-*"]

    types = ["astropy.modeling.mappings.Mapping"]

    def to_yaml_tree_transform(self, model, tag, ctx):
        node = {"mapping": list(model.mapping)}
        if model.n_inputs > max(model.mapping) + 1:
            node["n_inputs"] = model.n_inputs
        return node

    def from_yaml_tree_transform(self, node, tag, ctx):
        from astropy.modeling.models import Mapping

        return Mapping(tuple(node["mapping"]), node.get("n_inputs"))


class UnitsMappingConverter(Converter):
    """
    ASDF support for serializing the UnitsMapping model.
    Note that this converter does not inherit from TransformConverterBase,
    because the inputs and outputs are written differently
    from other models.
    """
    tags = ["tag:astropy.org:astropy/transform/units_mapping-*"]

    types = ["astropy.modeling.mappings.UnitsMapping"]

    def to_yaml_tree(self, model, tag, ctx):
        node = {}

        if model.name is not None:
            node["name"] = model.name

        inputs = []
        outputs = []
        for i, o, m in zip(model.inputs, model.outputs, model.mapping):
            input = {
                "name": i,
                "allow_dimensionless": model.input_units_allow_dimensionless[i],
            }
            if m[0] is not None:
                input["unit"] = m[0]
            if model.input_units_equivalencies is not None and i in model.input_units_equivalencies:
                input["equivalencies"] = model.input_units_equivalencies[i]
            inputs.append(input)

            output = {
                "name": o,
            }
            if m[-1] is not None:
                output["unit"] = m[-1]
            outputs.append(output)

        node["inputs"] = inputs
        node["outputs"] = outputs

        return node

    def from_yaml_tree(self, node, tag, ctx):
        from astropy.modeling.mappings import UnitsMapping

        mapping = tuple((i.get("unit"), o.get("unit"))
                        for i, o in zip(node["inputs"], node["outputs"]))

        equivalencies = None
        for i in node["inputs"]:
            if "equivalencies" in i:
                if equivalencies is None:
                    equivalencies = {}
                equivalencies[i["name"]] = i["equivalencies"]

        kwargs = {
            "input_units_equivalencies": equivalencies,
            "input_units_allow_dimensionless": {
                i["name"]: i.get("allow_dimensionless", False) for i in node["inputs"]},
        }

        if "name" in node:
            kwargs["name"] = node["name"]

        return UnitsMapping(mapping, **kwargs)
